/*
 * Copyright 1997 Vectaport Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/maproute.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapviewstate.h>
#include <OverlayUnidraw/ovviewer.h>
#include <UniIdraw/idarrows.h>
#include <Unidraw/Graphic/util.h>
#include <Unidraw/Graphic/verts.h>
#include <Unidraw/Tools/tool.h>
#include <Unidraw/catalog.h>
#include <Unidraw/clipboard.h>
#include <Unidraw/manips.h>
#include <Unidraw/unidraw.h>
#include <Attribute/paramlist.h>
#include <InterViews/event.h>
#include <InterViews/transformer.h>
#include <IV-2_6/InterViews/rubcurve.h>
#include <IV-2_6/InterViews/rubverts.h>
#include <Map/mapelt.h>
#include <Map/mapprim.h>
#include <OS/math.h>
#include <iostream.h>

/****************************************************************************/

ParamList* MapRouteComp::_maproute_params = nil;

MapRouteComp::MapRouteComp(MapElement* me)
  : MapFeatureComp(me) {
}

MapRouteComp::MapRouteComp(istream& in, OverlayComp* parent) : 
  MapFeatureComp(nil, parent)
{
    _valid = GetParamList()->read_args(in, this);
}

ParamList* MapRouteComp::GetParamList() {
    if (!_maproute_params) 
	GrowParamList(_maproute_params = new ParamList());
    return _maproute_params;
}

void MapRouteComp::GrowParamList(ParamList* pl) {
    pl->add_param("points", ParamStruct::required, &MapRouteScript::ReadPoints,
		  this, &_feature);
    MapFeatureComp::GrowParamList(pl);
    return;
}

MapRouteComp::~MapRouteComp() {}

Component* MapRouteComp::Copy() {
    MapRouteComp* mfc = new MapRouteComp((MapElement*)_feature);
    mfc->graphic(_gr);
    return mfc;
}

ClassId MapRouteComp::GetClassId () { return MAP_ROUTE_COMP; }

boolean MapRouteComp::IsA (ClassId id) {
    return MAP_ROUTE_COMP == id || MapFeatureComp::IsA(id);
}


/****************************************************************************/

MapRouteView::MapRouteView(MapRouteComp* comp) : MapFeatureView(comp)
{

}

MapRouteView::~MapRouteView() {}

ClassId MapRouteView::GetClassId () { return MAP_ROUTE_VIEW; }

boolean MapRouteView::IsA (ClassId id) {
    return MAP_ROUTE_VIEW == id || MapFeatureView::IsA(id);
}

Graphic* MapRouteView::HighlightGraphic() {
  return nil;
}

void MapRouteView::GetVertices (Coord*& x, Coord*& y, int& n) {
    Vertices* vertices = (Vertices*) GetGraphic();
    Transformer t;
    const Coord* origx, *origy;

    n = vertices->GetOriginal(origx, origy);
    ArrayDup(origx, origy, n, x, y);
    vertices->TotalTransformation(t);
    t.TransformList(x, y, n);
}

void MapRouteView::CreateHandles () {
    Coord* x, *y;
    int n;
    Viewer* v = GetViewer();
    
    if (v != nil) {
        GetVertices(x, y, n);
        _handles = new RubberHandles(nil, nil, x, y, n, 0, HANDLE_SIZE);
        v->InitRubberband(_handles);
        delete x;
        delete y;
    }
}

Graphic* MapRouteView::GetGraphic () {
    Graphic* graphic = GraphicView::GetGraphic();
    if (graphic != nil) return graphic;

    MapRouteComp* comp = (MapRouteComp*)GetGraphicComp();
    MapElement* feature = (MapElement*)comp->feature();
    Graphic* gr = nil; 

    Graphic* compgr = comp->GetGraphic();
    FullGraphic gs(compgr ? compgr : nil);
    gs.SetPattern(psnonepat);
    Catalog* catalog = unidraw->GetCatalog();
    PSBrush* br = catalog->FindBrush(0xffff, 2);
    gs.SetBrush(br);
    PSColor* col = catalog->FindColor("Red");
    gs.SetColors(col, col);

    double* x, *y;
    int n = feature->points(x, y);
    Coord* ix = new Coord[n];
    Coord* iy = new Coord[n];
    for (int i = 0; i < n; i++) {
      ix[i] = Math::round(x[i]);
      iy[i] = Math::round(y[i]);
    }
    gr = new ArrowMultiLine(ix, iy, n, false, false, 1.0, &gs);
    SetGraphic(gr);
    comp->graphic(gr);
    gr->SetTag(this);
    return gr;
}

void MapRouteView::Update() {
    MapRouteComp* comp = (MapRouteComp*)GetGraphicComp();
    MapElement* element = (MapElement*)comp->feature();
    ComponentView* topview = this;
    while (!topview->IsA(MAP_IDRAW_VIEW) && topview->GetParent())
      topview = topview->GetParent();
    MapEditor* mapeditor = ((MapIdrawView*)topview)->GetMapEditor();
    MapViewState* mapviewstate = mapeditor->GetMapViewState();
    
    Graphic* gr = GetGraphic();  
    gr->SetTransformer(new Transformer(mapviewstate->deftrans()));
    
    double *x, *y;
    int npts = element->points(x, y);
    int* ix = new int[npts];
    int* iy = new int[npts];
    for (int i = 0; i < npts; i++) {
      ix[i] = Math::round(x[i]);
      iy[i] = Math::round(y[i]);
    }
    IncurDamage(gr);
    MultiLineObj* ml = new MultiLineObj(ix, iy, npts);
    ((ArrowMultiLine*)gr)->SetOriginal(ml);
    ((ArrowMultiLine*)gr)->SetArrows(false, true);
    ((ArrowMultiLine*)gr)->ScaleArrows(5.0);
    IncurDamage(gr);
}


/*****************************************************************************/

MapRouteScript::MapRouteScript (MapRouteComp* subj) 
: MapFeatureScript(subj) {}

ClassId MapRouteScript::GetClassId () { return MAP_ROUTE_SCRIPT; }

boolean MapRouteScript::IsA (ClassId id) {
    return MAP_ROUTE_SCRIPT == id || MapFeatureScript::IsA(id);
}

boolean MapRouteScript::Definition (ostream& out) {
    MapRouteComp* comp = (MapRouteComp*)GetSubject();
    double* x, *y;
    MapElement* element = (MapElement*)comp->feature();
    int n = element->points(x, y);
    out << "route(";
    const int linemax = 5;
    for (int i=0; i<n; i+=linemax) {
        if (i) { 
	    out << ",\n"; 
	    Indent(out); 
	    out << "  ";
	}
        for (int j=i; j<i+linemax && j<n; j++) {
	    if (j!=i) out << ",";
	    out.form( "%.6f,%.6f", x[j], y[j] );
        }
    }
    Attributes(out);
    out << ")\n";
    
    return out.good();
}    


int MapRouteScript::ReadPoints (istream& in, void* addr1, 
    void* addr2, void* addr3, void* addr4) {
    double* x, *y;
    int n, bad;

    char ch = in.peek();
    if (ch != ')' && ch != ':') 
	bad = ParamList::parse_dblpts(in, x, y, n);
    else {
	x = y = nil;
	n = 0;
	bad = 0;
    }

    if (!in.good() || bad) {
	delete x;
	delete y;
        cerr << "abnormal exit from MapRouteScript::ReadPoints\n";
	return -1;
    }
    else {
        *(MapElement**)addr1 = new MapElement(0, MapFeature::LineType, 
					      n, x, y);
	delete x;
	delete y;
        return 0;
    }
}

/****************************************************************************/

RouteTool::RouteTool(ControlInfo* ci) : Tool(ci) {}

Manipulator* RouteTool::CreateManipulator(
    Viewer* v, Event& e, Transformer* rel
) {
    Manipulator* m = nil;

        v->Constrain(e.x, e.y);
        Coord x[1], y[1];
        x[0] = e.x;
        y[0] = e.y;
        GrowingVertices* rub = new GrowingMultiLine(
            nil, nil, x, y, 1, -1, HANDLE_SIZE
        );
#if 0
	if (((OverlayViewer*)v)->scribble_pointer()) 
            m = new ScribbleVertexManip(
	        v, rub, rel, this, DragConstraint(HorizOrVert | Gravity)
	    );
	else
#endif
            m = new VertexManip(
	        v, rub, rel, this, DragConstraint(HorizOrVert | Gravity)
	    );

    return m;
}

Command* RouteTool::InterpretManipulator(Manipulator* m) {
    DragManip* dm = (DragManip*) m;
    MapEditor* ed = (MapEditor*)dm->GetViewer()->GetEditor();
    Transformer* rel = dm->GetTransformer();
    Command* cmd = nil;
    Transformer* trans = ed->GetMapViewState()->deftrans();


        GrowingVertices* gv = (GrowingVertices*) dm->GetRubberband();
        Coord* x, *y;
        int n, pt;
        gv->GetCurrent(x, y, n, pt);

        if (n > 2 || x[0] != x[1] || y[0] != y[1]) {
            if (rel != nil) {
                rel = new Transformer(rel);
                rel->Invert();
            }
	    float* dx = new float[n];
	    float* dy = new float[n];
	    float* utmx = new float[n];
	    float* utmy = new float[n];
	    for (int i = 0; i < n; i++)
	      rel->transform(x[i], y[i], dx[i], dy[i]);
	    if (trans) {
	      for (int j = 0; j < n; j++)
		trans->inverse_transform(dx[j], dy[j], utmx[j], utmy[j]);
	      delete [] dx;
	      delete [] dy;
	    }
	    else {
	      delete [] utmx;
	      delete [] utmy;
	      utmx = dx;
	      utmy = dy;
	    }
	    double* ex = new double[n];
	    double* ey = new double[n];
	    for (int i = 0; i < n; i++) {
	      ex[i] = utmx[i];
	      ey[i] = utmy[i];
	    }
	    delete [] utmx;
	    delete [] utmy;
	    MapElement* route = new MapElement(0, MapFeature::LineType,
					       n, ex, ey);
	    cmd = new PasteCmd(ed, new Clipboard(new MapRouteComp(route)));
        }
        delete x;
        delete y;
    return cmd;
}

/****************************************************************************/

PickRouteTool::PickRouteTool(ControlInfo* ci) : MapInspectTool(nil, ci) {}

void PickRouteTool::DoInspect(GraphicView* view, Viewer* v) {
  if (view == nil) 
    return;
  GraphicComp* grcomp = view->GetGraphicComp();
  if (grcomp->IsA(MAPFEATURE_COMP)) {
    MapFeatureComp* mfc = (MapFeatureComp*)grcomp;
    MapFeature* mf = mfc->feature();
    if (mf->type() == MapFeature::LineType) {
      MapEdge* edge = mf->edge_primitive();

      double* xpts = new double[edge->npts()];
      double* ypts = new double[edge->npts()];
      for (int i=0; i<edge->npts(); i++) {
	xpts[i] = edge->xpoints()[i];
	ypts[i] = edge->ypoints()[i];
      }
      MapElement* route = 
	new MapElement(0, MapFeature::LineType,
		       edge->npts(), xpts, ypts);
      PasteCmd* cmd = new PasteCmd(v->GetEditor(),
				   new Clipboard(new MapRouteComp(route)));
      cmd->Execute();
    }
  }
}


